@@ -1,6 +1,5 @@
-// edition:2021
+//@ edition:2021
 
-#![feature(async_fn_in_trait)]
 #![allow(incomplete_features)]
 
 pub trait Foo {
